import sys

try:
    import ldap
    import ldap.async
    import ldap.modlist
except ImportError:
    pass

import lasso

from quixote import get_session, get_publisher

from qommon.form import *
from qommon.storage import StorableObject

from authentic import identities

import directory

class IdentitiesStorePratic:
    label = N_('Pr@tic Directory')
    admin_keys = ('pratic_ldap_url', 'pratic_ldap_base')

    def is_bootstrapping(self):
        return False

    def fill_admin_form(self, form, data_source):
        form.add(StringWidget, 'pratic_ldap_url', title = _('LDAP URL'), required = True,
                hint = htmltext(_('Example: <tt>ldap://directory.example.com</tt>')),
                value = data_source.get('pratic_ldap_url', ''))
        form.add(StringWidget, 'pratic_ldap_base', title = _('LDAP Base'), required = True,
                hint = htmltext(_('Example: <tt>dc=example, dc=com</tt>')),
                value = data_source.get('pratic_ldap_base', ''))

    def has_identity(self, uid):
        try:
            self.get_identity(uid)
        except KeyError:
            return False
        return True

    def load_identities(self):
        pass


    def get_identity_for_account(self, username, password, collectivity):
        store = directory.get_store()
        ldap_conn = ldap.initialize(store.pratic_ldap_url)
        # python-ldap only accept UTF8 encoded string so to pass string to python-ldap
        # API you must always .encode('utf8') them.
        # 
        # String coming from our own application are encoded with
        # get_publisher().site_charset, so you must decode them before trying to use
        # them.
        uid = 'uid=%s,%s' % (ldap.dn.escape_dn_chars(username.decode(get_publisher().site_charset)), collectivity)
        try:
            ldap_conn.simple_bind_s(uid.encode('utf8'), password.decode(get_publisher().site_charset).encode('utf8'))
        except ldap.INVALID_CREDENTIALS:
            return None

        identity = self.get_identity(uid.encode(get_publisher().site_charset))
        if identity:
            agent = identity.get_as_agent()
            if agent.cdg59isDisabled:
                return None
        return identity
    
    def get_identity_for_name_identifier(self, name_identifier):
        for i in MiniIdentityPratic.values():
            if not (i.lasso_dump or i.lasso_proxy_dump):
                continue
            if not ((name_identifier in (i.lasso_dump or '')) or (
                        name_identifier in (i.lasso_proxy_dump or ''))):
                continue # shortcut so it doesn't do the whole thing if not necessary
            identity = lasso.Identity.newFromDump(i.lasso_dump)
            for p in identity.providerIds:
                federation = identity.getFederation(p)
                if federation.localNameIdentifier and \
                        federation.localNameIdentifier.content == name_identifier:
                    return self.get_identity(i.id)
                if federation.remoteNameIdentifier and \
                        federation.remoteNameIdentifier.content == name_identifier:
                    return self.get_identity(i.id)

            if i.lasso_proxy_dump:
                identity = lasso.Identity.newFromDump(i.lasso_proxy_dump)
                for p in identity.providerIds:
                    federation = identity.getFederation(p)
                    if federation.localNameIdentifier and \
                            federation.localNameIdentifier.content == name_identifier:
                        return i
                    if federation.remoteNameIdentifier and \
                            federation.remoteNameIdentifier.content == name_identifier:
                        return i

        return None

    def init_session(self, session, password):
        session.pratic_ldap_password = password
        self._ldap_conn = None

    _ldap_conn = None
    def get_ldap_conn(self):
        if self._ldap_conn:
            return self._ldap_conn

        store = directory.get_store()
        self._ldap_conn = ldap.initialize(store.pratic_ldap_url)

        session = get_session()
        if session and session.user and getattr(session, 'pratic_ldap_password', None):
            uid = session.user
            password = session.pratic_ldap_password
            try:
                self._ldap_conn.simple_bind_s(uid, password)
            except ldap.INVALID_CREDENTIALS:
                pass # XXX?

        return self._ldap_conn
    ldap_conn = property(get_ldap_conn)

    def connect(self, session):
        if session and session.user and getattr(session, 'pratic_ldap_password', None):
            # bind to the LDAP directory so it can access its attributes
            try:
                self.ldap_conn.simple_bind_s(session.user,
                        session.pratic_ldap_password)
            except ldap.INVALID_CREDENTIALS:
                return

    def get_identity(self, uid):
        try:
            identity = MiniIdentityPratic.get(uid)
        except KeyError:
            identity = MiniIdentityPratic(id = uid)
        return identity

    def save(self, identity):
        identity.store()


class MiniIdentityPratic(StorableObject):
    _names = 'pratic_identities'
    lasso_dump = None
    lasso_proxy_dump = None
    resource_id = None

    roles = None
    accounts = None

    def get_name(self):
        return str(self.id).split(',')[0].split('=')[1]
    name = property(get_name)

    def get_locked_down(self):
        return False
    locked_down = property(get_locked_down)

    def get_disabled(self):
        agent = self.get_as_agent()
        if agent.cdg59isDisabled:
            return True
        return False
    disabled = property(get_disabled)

    def register_new_session(self):
        agent = self.get_as_agent()
        session = get_session()
        session.previousConnectionTime = agent.cdg59lastConnectionTime
        session.previousConnectionDuration = agent.cdg59lastConnectionDuration
        agent.cdg59lastConnectionTime = str(int(time.time()))
        agent.cdg59lastConnectionDuration = None
        agent.save(['cdg59lastConnectionTime', 'cdg59lastConnectionDuration'])

    def is_admin(self):
        return ',ou=admin,' in str(self.id)

    def is_local_admin(self):
        if not 'ou=' in str(self.id):
            return False
        store = directory.get_store()
        ldap_conn = ldap.initialize(store.pratic_ldap_url)
        try:
            record = ldap_conn.search_s(str(self.id), ldap.SCOPE_BASE)[0][1]
        except:
            return False
        return 'TRUE' in record.get('cdg59isAdmin', [])

    def get_display_name(self):
        return self.id[:self.id.index(',')]
    display_name = property(get_display_name)
    
    def get_collectivity(self):
        collectivity_ou = self.id[self.id.index(',')+1:]
        collectivity_ou = collectivity_ou[:collectivity_ou.index(',')].strip().split('=')[1]
        return directory.get_collectivity(collectivity_ou)

    def get_attributes(self):
        agent = self.get_as_agent()
        attributes = {}
        for k in ('username', 'cn', 'mail'):
            v = getattr(agent, k)
            if v is None:
                continue
            attributes[k] = [ unicode(v, 'iso-8859-1'), ]
        if agent.cdg59isAdmin:
            attributes['local-admin'] = [ 'true', ]
        if self.is_admin():
            attributes['super-admin'] = [ 'true', ]
        else:
            attributes['collectivity'] = [ unicode(self.get_collectivity().ou, 'iso-8859-1'), ]
        return attributes
    attributes = property(get_attributes)

    _agent = None
    def get_as_agent(self):
        if self._agent:
            return self._agent
        store = directory.get_store()
        ldap_conn = ldap.initialize(store.pratic_ldap_url)
        record = ldap_conn.search_s(self.id, ldap.SCOPE_BASE)[0]
        self._agent = directory.Agent(record)
        return self._agent

    def __str__(self):
        # used for logging
        if type(self.id) is not str:
            return ''
        if not self.id:
            return ''
        if not '=' in self.id:
            # logger system could make self.id be 'unlogged'...
            return self.id
        return '/'.join([x.split('=', 1)[1] for x in self.id.split(',')[:2]])

    # don't pickle _agent cache
    def __getstate__(self):
        odict = self.__dict__
        if odict.has_key('_agent'):
            del odict['_agent']
        return odict

    def __setstate__(self, dict):
        self.__dict__ = dict
        self._agent = None


identities.stores['pratic'] = IdentitiesStorePratic

